/*
  Copyright 1999 The University of Rhode Island and The Massachusetts
  Institute of Technology

  Portions of this software were developed by the Graduate School of
  Oceanography (GSO) at the University of Rhode Island (URI) in collaboration
  with The Massachusetts Institute of Technology (MIT).

  Access and use of this software shall impose the following obligations and
  understandings on the user. The user is granted the right, without any fee
  or cost, to use, copy, modify, alter, enhance and distribute this software,
  and any derivative works thereof, and its supporting documentation for any
  purpose whatsoever, provided that this entire notice appears in all copies
  of the software, derivative works and supporting documentation.  Further,
  the user agrees to credit URI/MIT in any publications that result from the
  use of this software or in any product that includes this software. The
  names URI, MIT and/or GSO, however, may not be used in any advertising or
  publicity to endorse or promote any products or commercial entity unless
  specific written permission is obtained from URI/MIT. The user also
  understands that URI/MIT is not obligated to provide the user with any
  support, consulting, training or assistance of any kind with regard to the
  use, operation and performance of this software nor to provide the user
  with any updates, revisions, new versions or "bug fixes".

  THIS SOFTWARE IS PROVIDED BY URI/MIT "AS IS" AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
  EVENT SHALL URI/MIT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
  DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS
  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE ACCESS, USE OR PERFORMANCE
  OF THIS SOFTWARE.
*/
// DODS-netCDF surrogate library functions for file handling. Note that with
// the preprocessor symbol LOCAL defined this library will access local files
// without using the DODS network services. Defining LOCAL will add support
// for local files through a slightly modifed version of the original netCDF
// library. This library has the same functionality as the original (it can
// access only files on a local hard disk) but has renamed entry points to
// avoid linker conflicts.
// 
// ReZa 10/13/94
//
// Rewrite to support netCDF version 3 interface; 5/12/99, Reza
//
// $Log: Dnc.cc,v $
// Revision 1.1  1999/07/28 00:22:41  jimg
// Added
//

#include "config_dap.h"

static char rcsid[] not_used ={"$Id: Dnc.cc,v 1.1 1999/07/28 00:22:41 jimg Exp $"};

#ifdef __GNUG__
#pragma implementation
#endif

#include <assert.h>
#include <string>

#include "NCStructure.h"
#include "Dnetcdf.h"
#include "Connect.h"
#include "NCConnect.h"

// Make these symbols external so as to avoid link-time conflicts with their
// definitions in the localized version of the original
// implementation of netcdf.

#ifdef LOCAL
extern char *cdf_routine_name;
extern int ncopts;
extern int ncerr;
#else
char *cdf_routine_name;
int ncopts = NC_VERBOSE;
int ncerr;
#endif

int rcode;

Connections<NCConnect *> *conns = 0;


int
nc__create(const char * path, int ioflags, size_t initialsz,
        size_t *chunksizehintp, int *ncid_ptr)
{
    int id;
    NCConnect *c = new NCConnect(path);

#ifdef LOCAL
    if(c->is_local()) {  
	rcode = lnc__create(path, ioflags, initialsz,
			    chunksizehintp, ncid_ptr);
	if (rcode == NC_NOERR) {
	    c->ncid() = *ncid_ptr;
	    if (!conns)
		conns = new Connections<NCConnect *>(MAX_NC_OPEN);
	    *ncid_ptr = conns->add_connect(c);
	    return NC_NOERR;
	}
	else {
	  //tmp_out	    delete c;
	    return rcode;
	}
    }
#endif

    //tmp_out    delete c;
    return NC_EPERM;
}


int
nc_create(const char * path, int ioflags, int *ncid_ptr)
{
        return nc__create(path, ioflags, 0, NULL, ncid_ptr);
}


int
nccreate(const char *path, int cmode)
{
    int id;
    NCConnect *c = new NCConnect(path);

#ifdef LOCAL
    if(c->is_local()) {  
	rcode = lnccreate(path, cmode);
	if (rcode != -1) {
	    c->ncid() = rcode;
	    if (!conns)
		conns = new Connections<NCConnect *>(MAX_NC_OPEN);
	    id = conns->add_connect(c);
	    return id;
	}
	else {
	  //tmp_out	    delete c;
	    return rcode;
	}
    }
#endif

    cdf_routine_name = "nccreate" ;
    NCadvise(NC_EPERM, "DODS network interface is read-only") ;

    //tmp_out    delete c;
    return(-1);
}


nc__open(const char * urlpath, int ioflags, size_t *chunksizehintp, int *ncid_ptr)
{
    int id;
    string path = urlpath;
    bool status = true;
    string var_name;

    size_t q_mark = path.find("?");
    string ce = "";
    if (q_mark != string::npos) {
	ce = path.substr(q_mark + 1);
	path = path.substr(0, q_mark);
    }

    NCConnect *c = new NCConnect(path);

#ifdef LOCAL
    if (c->is_local()) {  
	rcode = lnc__open(path.c_str(), ioflags, chunksizehintp, ncid_ptr);
	if (rcode == NC_NOERR) {
	    c->ncid() = *ncid_ptr;
	    if (!conns)
		conns = new Connections<NCConnect *>(MAX_NC_OPEN);
	    *ncid_ptr = conns->add_connect(c);
	    return NC_NOERR;
	}
	else {
	  //tmp_out	    delete c;
	    return rcode;
	}
    }
#endif

    if (ioflags != NC_NOWRITE){
      //tmp_out	delete c;
      return NC_EPERM;
    }

    if (!c->request_das()){
      //tmp_out	delete c;
      return NC_EINVAL;
    }

    if (!c->request_dds()){ 
      //tmp_out	delete c;
      return NC_EINVAL;
    }

    if (!c->dds().parse_constraint(ce)) {
      //tmp_out	delete c;
      return NC_EINVAL;
    }

    if (!c->request_Cdds(ce)) {
      //tmp_out	delete c;
      return NC_EINVAL;
    }

    // Remove all the non-supported types for now
    Pix Cp=c->dds().first_var();
    //    if((c->dds().var(Cp)->type() != dods_grid_c )&& 
    //   (c->dds().var(Cp)->type() != dods_array_c )){ 
    if((c->dds().var(Cp)->type() == dods_structure_c ) ||
       (c->dds().var(Cp)->type() == dods_sequence_c ) || 
       (c->dds().var(Cp)->type() == dods_list_c )){ 
      var_name = c->dds().var(Cp)->name();
      c->dds().next_var(Cp);
      c->dds().del_var(var_name);
      c->Cdds().del_var(var_name);
      //c->das().del_attr(var_name);
    }
    else
      c->dds().next_var(Cp);

    while(Cp) {

      //      if((c->dds().var(Cp)->type() != dods_grid_c )&& 
      //	 (c->dds().var(Cp)->type() != dods_array_c )){ 
      if((c->dds().var(Cp)->type() == dods_structure_c ) ||
	 (c->dds().var(Cp)->type() == dods_sequence_c ) || 
	 (c->dds().var(Cp)->type() == dods_list_c )){ 
	
	var_name = c->dds().var(Cp)->name();
	c->dds().next_var(Cp);

	c->dds().del_var(var_name);
	c->Cdds().del_var(var_name);
	//c->das().del_attr(var_name);
      }
      else
	c->dds().next_var(Cp);	
    }

    // well now we can extract dim. and attr. info. into tables
    c->parse_das_loc();
    c->parse_dims();

    if (!conns)
      conns = new Connections<NCConnect *>(MAX_NC_OPEN);

    *ncid_ptr = conns->add_connect(c);
    return NC_NOERR;
 
}


int
nc_open(const char * path, int ioflags, int *ncid_ptr)
{
        return nc__open(path, ioflags, NULL, ncid_ptr);
}


int
ncopen(const char *urlpath, int cmode)
{
    int id;
    string path = urlpath;
    bool status = true;
    string var_name;

    size_t q_mark = path.find("?");
    string ce = "";
    if (q_mark != string::npos) {
	ce = path.substr(q_mark + 1);
	path = path.substr(0, q_mark);
    }

    NCConnect *c = new NCConnect(path);

#ifdef LOCAL
    if (c->is_local()) {  
	rcode = lncopen(path.c_str(), cmode);
	if (rcode != -1) {
	    c->ncid() = rcode;
	    if (!conns)
		conns = new Connections<NCConnect *>(MAX_NC_OPEN);
	    id = conns->add_connect(c);
	    return id;
	}
	else {
	  //tmp_out	    delete c;
	    return rcode;
	}
    }
#endif

    cdf_routine_name = "ncopen";

    if (cmode != NC_NOWRITE) {    
	NCadvise(NC_EPERM, 
		 "Bad file mode, Only read access allowed over the network");
	return -1;
	//	status = false;
	//	goto exit;
    }

    if (!c->request_das()) {
	NCadvise(NC_EINVAL, "Error in getting DAS from `%s'", path.c_str());
	return -1;
    }

    if (!c->request_dds()) {
	NCadvise(NC_EINVAL, "Error in getting DDS from `%s'", path.c_str());
	return -1;
    }

    if (!c->dds().parse_constraint(ce)) {
	NCadvise(NC_EINVAL, "Error in parsing constarint from `%s'",
		 path.c_str());
	return -1;
    }

    if (!c->request_Cdds(ce)) {
	NCadvise(NC_EINVAL, "Error in getting CDDS from `%s'", path.c_str());
	return -1;
    }

#ifdef TEST
    // Use variable rename to resolve structures
    for (Pix Sp=c->dds().first_var(); Sp; c->dds().next_var(Sp))
      if (c->dds().var(Sp)->type() == dods_structure_c){
	NCStructure *stq = (NCStructure *) c->dds().var(Sp);
	for (Pix p=stq->first_var(); p!=0; stq->next_var(p))
	  if ((stq->var(p)->type() == dods_grid_c) ||
	      (stq->var(p)->type() == dods_array_c)){
	    //
	    stq->var(p)->set_name(stq->name() + "." + stq->var(p)->name()); 

	    //
	    c->dds().add_var(stq->var(p));
	    c->Cdds().add_var(stq->var(p));
	  }
      }

    c->dds().print();
#endif

    Pix Cp=c->dds().first_var();
    if((c->dds().var(Cp)->type() != dods_grid_c )&& 
       (c->dds().var(Cp)->type() != dods_array_c )){ 
      var_name = c->dds().var(Cp)->name();
      c->dds().next_var(Cp);
      c->dds().del_var(var_name);
      c->Cdds().del_var(var_name);
      //c->das().del_attr(var_name);
    }
    else
      c->dds().next_var(Cp);

    while(Cp) {

      if((c->dds().var(Cp)->type() != dods_grid_c )&& 
	 (c->dds().var(Cp)->type() != dods_array_c )){ 
	
	var_name = c->dds().var(Cp)->name();
	c->dds().next_var(Cp);

	c->dds().del_var(var_name);
	c->Cdds().del_var(var_name);
	//c->das().del_attr(var_name);
      }
      else
	c->dds().next_var(Cp);
	
    }

#ifdef TEST
    //    c->das().print();
    c->dds().print();
    c->Cdds().print();
#endif

    // well now we can extract dim. and attr. info. into tables
    c->parse_das_loc();
    c->parse_dims();

#ifdef TEST
    for (int i=0 ; i<7 ; i++)
	cerr << i << "-" << c->das_loc(i) << endl;
    c->das().print();    c->dds().next_var(Cp);
    c->dds().print();
#endif

exit:
    if (status == true) {
	if (!conns)
	    conns = new Connections<NCConnect *>(MAX_NC_OPEN);

	return(conns->add_connect(c)) ;
    }
    else {
      //tmp_out	delete c;
	return -1;
    }
}


int
nc_close(int cdfid)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
      return NC_EBADID;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lnc_close((*conns)[cdfid]->ncid());

	delete (*conns)[cdfid];
	conns->del_connect(cdfid);

	return rcode;
    }
#endif

#ifdef TEST
    (*conns)[cdfid]->das().print();
    (*conns)[cdfid]->dds().print();
#endif

    delete (*conns)[cdfid];

    conns->del_connect(cdfid);
    
    return NC_NOERR;
}


int
ncclose(int cdfid)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;
#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lncclose((*conns)[cdfid]->ncid());

	delete (*conns)[cdfid];
	conns->del_connect(cdfid);

	return rcode;
    }
#endif

    cdf_routine_name = "ncclose" ;

    delete (*conns)[cdfid];

    conns->del_connect(cdfid);
    
    return(0) ;
}


extern int num_attr(int cdfid, int varid);

int
nc_inq(int cdfid, int *ndimsp, int *nvarsp, int *nattsp, int *xtendimp)
{
 if((*conns)[cdfid] == NULL)
   return NC_EBADID;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lnc_inq((*conns)[cdfid]->ncid(), ndimsp, nvarsp, nattsp, xtendimp);
	return rcode;
    }
#endif

  if(nvarsp != NULL)
    *nvarsp = (*conns)[cdfid]->nvars();

  if(nattsp != NULL)
    *nattsp = num_attr(cdfid,NC_GLOBAL);

  if(ndimsp != NULL)
    *ndimsp = (*conns)[cdfid]->ndims();
		
  if(xtendimp != NULL)
    *xtendimp = (*conns)[cdfid]->recdim(); 
	
  return NC_NOERR;
}


int 
nc_inq_ndims(int cdfid, int *ndimsp)
{
  if((*conns)[cdfid] == NULL)
    return NC_EBADID;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lnc_inq_ndims((*conns)[cdfid]->ncid(), ndimsp);
	return rcode;
    }
#endif

  if(ndimsp != NULL)
    *ndimsp = (*conns)[cdfid]->ndims();
		
  return NC_NOERR;
}


int 
nc_inq_nvars(int cdfid, int *nvarsp)
{
  if((*conns)[cdfid] == NULL)
    return NC_EBADID;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lnc_inq_nvars((*conns)[cdfid]->ncid(), nvarsp);
	return rcode;
    }
#endif

  if(nvarsp != NULL)
    *nvarsp = (*conns)[cdfid]->nvars();

  return NC_NOERR;
}


int 
nc_inq_natts(int cdfid, int *nattsp)
{
  if((*conns)[cdfid] == NULL)
    return NC_EBADID;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lnc_inq_natts((*conns)[cdfid]->ncid(), nattsp);
	return rcode;
    }
#endif

  if(nattsp != NULL)
    *nattsp = num_attr(cdfid,NC_GLOBAL);

  return NC_NOERR;
}


int 
nc_inq_unlimdim(int cdfid, int *xtendimp)
{
  if((*conns)[cdfid] == NULL)
    return NC_EBADID;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lnc_inq_unlimdim((*conns)[cdfid]->ncid(), xtendimp);
	return rcode;
    }
#endif

  if(xtendimp != NULL)
    *xtendimp = (*conns)[cdfid]->recdim(); 
	
  return NC_NOERR;
}


int
ncinquire(int cdfid,int *ndimsp,int *nvarsp, int *nattrsp, int *xtendimp)
{
  if((*conns)[cdfid] == NULL) // does the file id exist?
    return(-1) ;

#ifdef LOCAL
  if((*conns)[cdfid]->is_local()) {
    rcode = lncinquire((*conns)[cdfid]->ncid(), ndimsp, nvarsp,
			      nattrsp,xtendimp);
    return rcode;
  }
#endif
  
  cdf_routine_name = "ncinquire" ;

  if(nvarsp != NULL)
    *nvarsp = (*conns)[cdfid]->nvars();

  if(nattrsp != NULL)
    *nattrsp = num_attr(cdfid,NC_GLOBAL);

  if(ndimsp != NULL)
    *ndimsp = (*conns)[cdfid]->ndims();
		
  if(xtendimp != NULL)
    *xtendimp = (*conns)[cdfid]->recdim(); 
	
  return(cdfid) ;
}


int
nc_sync(int cdfid)
{
    assert(conns); 

    if((*conns)[cdfid] == NULL) // does the file id exist?
      return NC_EBADID ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
      rcode = lnc_sync((*conns)[cdfid]->ncid());
      return rcode;
    }
#endif

    return NC_EPERM;
}


int
ncsync(int cdfid)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lncsync((*conns)[cdfid]->ncid());
	return rcode;
    }
#endif

    cdf_routine_name = "ncsync" ;
    NCadvise(NC_EPERM, "DODS network interface is read-only") ;

    return(-1);
}


int
nc_abort(int cdfid)
{
    assert(conns); 

    if((*conns)[cdfid] == NULL) // does the file id exist?
      return NC_EBADID ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
      rcode = lnc_abort((*conns)[cdfid]->ncid());
      return rcode;
    }
#endif
   rcode = nc_close(cdfid);
   return rcode;
}


int
ncabort(int cdfid)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
	rcode = lncabort((*conns)[cdfid]->ncid());
	return rcode;
    }
#endif
    
    cdf_routine_name = "ncabort" ;

    rcode = ncclose(cdfid);

    return rcode;
}


/* 
 * Deprecated function ;
 */
int
ncnobuf(int cdfid)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

    cdf_routine_name = "ncnobuf" ;

    /* NOOP */

    return(0);
}


int
nc_redef(int cdfid)
{
    if(!conns)
      return NC_EBADID; 

    if((*conns)[cdfid] == NULL) // does the file id exist?
      return NC_EBADID ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
      rcode = lnc_redef((*conns)[cdfid]->ncid());
      return rcode;
    }
#endif

    return NC_EPERM;
}

  
int
ncredef(int cdfid)
{
    if(!conns)
      return NC_EBADID;

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {  
	rcode = lncredef((*conns)[cdfid]->ncid());
	return rcode;
    }
#endif

    cdf_routine_name = "ncredef" ;
    NCadvise(NC_EPERM, "DODS network interface is read-only") ;
    return(-1);
}


int
nc__enddef(int cdfid, size_t h_minfree, size_t v_align,
        size_t v_minfree, size_t r_align)
{
    assert(conns); 

    if((*conns)[cdfid] == NULL) // does the file id exist?
      return NC_EBADID ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
      rcode = lnc__enddef((*conns)[cdfid]->ncid(), h_minfree,
			  v_align, v_minfree, r_align);
      return rcode;
    }
#endif

    return NC_NOERR; // no effect on a remote file

}


int
nc_enddef(int cdfid)
{
    assert(conns); 

    if((*conns)[cdfid] == NULL) // does the file id exist?
      return NC_EBADID ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
      rcode = lnc_enddef((*conns)[cdfid]->ncid());
      return rcode;
    }
#endif

    return NC_NOERR; // no effect on a remote file
}


int
ncendef(int cdfid )
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?

	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) { 
	rcode = lncendef((*conns)[cdfid]->ncid());
	return rcode;
    }
#endif

    cdf_routine_name = "ncendef" ;

    return(0);
}


int
nc_set_fill(int cdfid, int fillmode, int *old_mode_ptr)
{
    assert(conns); 

    if((*conns)[cdfid] == NULL) // does the file id exist?
      return NC_EBADID ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {
      rcode = lnc_set_fill((*conns)[cdfid]->ncid(), fillmode, old_mode_ptr);
      return rcode;
    }
#endif

    return NC_EPERM;
}


int
ncsetfill(int cdfid, int fillmode)
{
    assert(conns);

    if((*conns)[cdfid] == NULL) // does the file id exist?
	return(-1) ;

#ifdef LOCAL
    if((*conns)[cdfid]->is_local()) {  
	rcode = lncsetfill((*conns)[cdfid]->ncid(), fillmode);
	return rcode;
    }
#endif

    cdf_routine_name = "ncsetfill" ;
    NCadvise(NC_EPERM, "DODS network interface is read-only") ;

    return(-1);
}


const char *
nc_inq_libvers(void)
{
#ifdef LOCAL
  return lnc_inq_libvers(); // local netCDF library version (not DODS)
#endif
}


const char *
nc_strerror(int err)
{
#ifdef LOCAL
  return lnc_strerror(err); 
#endif
}

